/***********************************************************************/
/* File: bmp.c                                                         */
/*                                                                     */
/*                                                                     */
/*---------------------------------------------------------------------*/
/* Programmer: Nils Schmeisser                                         */
/*                                                                     */
/*---------------------------------------------------------------------*/
/* Revision date: 8.2.96                                               */
/*---------------------------------------------------------------------*/
/* Notes: Microsoft Windows Bitmap Speicherroutine                     */
/*                                                                     */
/*                                                                     */
/***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "bmp.h"
#ifndef __cplusplus
  #include "cppmacro.h"
#endif

#include <stdlib.h>
#include "JpegFile.h"
#include "bmp.h"

const int xPALETTE[40][3]={
                 { 255, 255, 255},
                 {   0,   0,   0},
                 { 255, 255,   0},
                 {  60, 251,  52},
                 { 199, 199, 199},
                 { 231, 231, 231},
                 { 109, 109, 109},
                 { 211, 211, 211},
                 { 239, 239, 239},
                 { 117, 117, 117},
                 { 178, 178, 178},
                 {  97, 162, 162},
                 { 186, 215, 215},
                 {  48,  81,  81},
                 {  81, 138, 138},
                 { 170, 170, 170},
                 { 251,  20,  65},
                 {   0,   8,   0},
                 {   0,  16,   0},
                 {   0,  24,   0},
                 {   0,  32,   0},
                 {   0,  40,   0},
                 {   0,  48,   0},
                 {   0,  56,   0},
                 {   0,  65,   0},
                 {   0,  73,   0},
                 {   0,  81,   0},
                 {   0,  89,   0},
                 {   0,  97,   0},
                 {   0, 105,   0},
                 {   0, 113,   0},
                 {   0, 121,   0},
                 {   0, 130,   0},
                 {   0, 138,   0},
                 {   0, 150,   0},
                 {   0, 158,   0},
                 {   0, 166,   0},
                 {   0, 174,   0},
                 {   0, 182,   0},
                 {   0, 190,   0}
  };

unsigned short int *get_palette(char *lutname) {
  int i;
  FILE *flut;
  unsigned short *clut;

  if (lutname!=NULL) {
    if (strlen(lutname)>0) {
      if ((flut=fopen(lutname,"r"))!=NULL) {
	clut=new unsigned short [768];
	fread(&clut[  0],256*2,1,flut);
	fread(&clut[256],256*2,1,flut);
	fread(&clut[512],256*2,1,flut);
	for (i=0;i<256;i++) {
	  clut[i    ]=clut[i    ] >> 8;
	  clut[i+256]=clut[i+256] >> 8;
	  clut[i+512]=clut[i+512] >> 8;
	}
	fclose(flut);
	return clut;
      } else {
	return NULL;
      }
    } else 
      return NULL;
  } else
    return NULL;
}

RGB_QUAD RGB(byte r,byte g,byte b,byte i=0) {
  RGB_QUAD rgb;

  rgb.Red  =r;
  rgb.Green=g;
  rgb.Blue =b;
  rgb.Int  =i;

  return rgb;
}

RGB_QUAD PixToRGB(int i,RGB_QUAD *palette) {
  RGB_QUAD rgb;

  rgb=palette[i];
  return rgb;
}

int RGBToPix(int red,int green,int blue,RGB_QUAD *palette,int psize) {
  int i,c;
  double dr,dg,db;
  double dist,min=256.0*256.0*3.0+1.0;

  c=0;
  for (i=0;i<psize;i++) {
    dr=(double)(palette[i].Red-red);
    dg=(double)(palette[i].Green-green);
    db=(double)(palette[i].Blue-blue);
    dist=dr*dr+dg*dg+db*db;
    if (dist<min) {
      min=dist;
      c=i;
    }
  }
  return c;
}

dword swap(dword w) {
  byte b[4];

  b[0]=(0x0ff000000 & w)/0x1000000;
  b[1]=(0x000ff0000 & w)/0x10000;
  b[2]=(0x00000ff00 & w)/0x100;
  b[3]=(0x0000000ff & w);
  return b[3]*0x1000000+b[2]*0x10000+b[1]*0x100+b[0];
}

void ExportToBMP(int w,int h,int bpp,byte** data,char *name,char* lutname) {
  FILE *f;
  bmp_header header;
  RGB_QUAD *palette;
  RGB_QUAD rgb;

  int x,y,i;
  unsigned short int *lut=NULL;
  byte *b=NULL;
  
  if (data!=NULL) {
    f=fopen(name,"w+");
    header.id1=0x42;
    header.id2=0x4d;

    header.bfSize=BITMAP_HEADER_SIZE-2;
    if (bpp==1)
      header.bfSize+=2*4+w*h/8;
    if (bpp==4)
      header.bfSize+=16*4+w*h/2;
    if (bpp==8)
      header.bfSize+=256*4+w*h;
    if (bpp==24)
      header.bfSize+=3*w*h;

    header.res1=0;
    header.res2=0;

    header.bfOffs=BITMAP_HEADER_SIZE;
    if (bpp==1)
      header.bfOffs+=2*4;
    if (bpp==4)
      header.bfOffs+=16*4;
    if (bpp==8)
      header.bfOffs+=256*4;

    header.biSize=BITMAP_INFO_SIZE;
    header.biWidth=w;
    header.biHeight=h;
    header.biPlanes=1;
    header.biBitCnt=bpp;
    header.biCompr=BI_RGB;
    if (header.biCompr==BI_RGB)
      header.biSizeIm=header.bfSize-header.bfOffs;
    header.biXPels=0x0b6d;
    header.biYPels=0x0b6d;
    if (bpp==1)
      header.biClrUsed=2;
    if (bpp==4)
      header.biClrUsed=16;
    if (bpp==8)
      header.biClrUsed=256;
    if (bpp==24)
      header.biClrUsed=256;
    header.biClrImp=header.biClrUsed;

    fwrite(&header,2,1,f);
    fwrite(&header.bfSize,4,1,f);
    fwrite(&header.res1,2,1,f);
    fwrite(&header.res2,2,1,f);
    fwrite(&header.bfOffs,4,1,f);
    fwrite(&header.biSize,4,1,f);
    fwrite(&header.biWidth,4,1,f);
    fwrite(&header.biHeight,4,1,f);
    fwrite(&header.biPlanes,2,1,f);
    fwrite(&header.biBitCnt,2,1,f);
    fwrite(&header.biCompr,4,1,f);
    fwrite(&header.biSizeIm,4,1,f);
    fwrite(&header.biXPels,4,1,f);
    fwrite(&header.biYPels,4,1,f);
    fwrite(&header.biClrUsed,4,1,f);
    fwrite(&header.biClrImp,4,1,f);

    palette=NULL;
    if (bpp==1)
      palette=new RGB_QUAD [2];
    if (bpp==4)
      palette=new RGB_QUAD [16];
    if (bpp==8)
      palette=new RGB_QUAD [256];
    if ((bpp==24) && (lutname!=NULL))
      palette=new RGB_QUAD [256];

    if ((lut=get_palette(lutname))!=NULL) {
      for (i=0;i<256;i++) {
        palette[i].Blue =(byte)lut[i+512]; 
        palette[i].Green=(byte)lut[i+256]; 
        palette[i].Red  =(byte)lut[i    ]; 
        palette[i].Int  =0;
      }
      delete(lut);
    } else {
      /* here we should somehow generate a nice standard lookup table */
      if (bpp==1) {
	palette[0]=RGB(0x00,0x00,0x00);
	palette[1]=RGB(0xff,0xff,0xff);
      }
      if (bpp==4) {
	palette[ 0]=RGB(0x00,0x00,0x00);
	palette[ 1]=RGB(0x00,0x00,0x80);
	palette[ 2]=RGB(0x00,0x80,0x00);
	palette[ 3]=RGB(0x00,0x80,0x80);
	palette[ 4]=RGB(0x80,0x00,0x00);     
	palette[ 5]=RGB(0x80,0x00,0x80);
	palette[ 6]=RGB(0x80,0x80,0x00);
	palette[ 7]=RGB(0x80,0x80,0x80);
	palette[ 8]=RGB(0xc0,0xc0,0xc0);
	palette[ 9]=RGB(0x00,0x00,0xff);
	palette[10]=RGB(0x00,0xff,0x00);
	palette[11]=RGB(0x00,0xff,0xff);
	palette[12]=RGB(0xff,0x00,0x00);
	palette[13]=RGB(0xff,0x00,0xff);
	palette[14]=RGB(0xff,0xff,0x00);
	palette[15]=RGB(0xff,0xff,0xff);
      }
      if ((bpp==8) || ((bpp==24) && (lutname!=NULL))) {
      }
    } 

    if (bpp==1)
      fwrite(palette,4,2,f);
    if (bpp==4)
      fwrite(palette,4,16,f);
    if ((bpp==8) || ((bpp==24) && (lutname!=NULL)))
      fwrite(palette,4,256,f);
    if (palette!=NULL)
      delete palette;

    if (bpp==1)
      b=new byte [w/8+1];
    if (bpp==4)
      b=new byte [w/2+1];
    if (bpp==24)
      b=new byte [w*3];

    for (y=h-1;y>=0;y--) {
      if (bpp==4) {
	for (x=0;x<w;x++) {
	  b[x/2]=data[y][x]*16+data[y][x+1];
	  x++;
	}
	fwrite(b,1,w/2,f);
      }
      if (bpp==8)
	fwrite(data[y],1,w,f);
      if (bpp==24) {
	if (palette!=NULL) {
	  for (x=0;x<w;x++) {
	    rgb=PixToRGB(data[y][x],palette);
	    b[3*x+2]=rgb.Red;
	    b[3*x+1]=rgb.Green;
	    b[3*x+0]=rgb.Blue;
	  }
	  fwrite(b,3,w,f);
	} else {
	  for (x=0;x<w;x++) {
	    b[3*x+0]=data[y][3*x+0];
	    b[3*x+1]=data[y][3*x+1];
	    b[3*x+2]=data[y][3*x+2];
	  }
	  fwrite(b,1,3*w,f);   /* a real true-color source */
	}
      }
    }
    if (b!=NULL)
      delete b;

    fclose(f);
  }
}

void ExportToAVSFLD(int w,int h,byte** data,char *name) {
  FILE *f;
  int i;
  
  f=fopen(name,"w+");
  fprintf(f,"# AVS field file\n");
  fprintf(f,"ndim = 2\n");
  fprintf(f,"dim1 = %d\n",h);
  fprintf(f,"dim2 = %d\n",w);
  fprintf(f,"nspace = 2\n");
  fprintf(f,"veclen = 1\n");
  fprintf(f,"data = byte\n");
  fprintf(f,"field = uniform\n");
  fprintf(f,"%c%c",0x0c,0x0c);

  for (i=0;i<h;i++) fwrite(data[i],1,w,f);
  fclose(f);
}

RGBRec **ReadBMP(char *fname,int *w,int *h,int *bpp) {
  FILE *f;
  bmp_header header;
  byte *buffer;
  int i,j,k,m,psize;
  RGB_QUAD *palette=NULL;
  RGB_QUAD rgb;
  RGBRec **data;

  *w=0;
  *h=0;
  f=fopen(fname,"r");
  if (f!=NULL) {

    fread(&header,2,1,f);
    fread(&header.bfSize,4,1,f);
    fread(&header.res1,2,1,f);
    fread(&header.res2,2,1,f);
    fread(&header.bfOffs,4,1,f);
    fread(&header.biSize,4,1,f);
    fread(&header.biWidth,4,1,f);
    fread(&header.biHeight,4,1,f);
    fread(&header.biPlanes,2,1,f);
    fread(&header.biBitCnt,2,1,f);
    fread(&header.biCompr,4,1,f);
    fread(&header.biSizeIm,4,1,f);
    fread(&header.biXPels,4,1,f);
    fread(&header.biYPels,4,1,f);
    fread(&header.biClrUsed,4,1,f);
    fread(&header.biClrImp,4,1,f);

    *bpp=header.biBitCnt;
    *w=header.biWidth;
    *h=header.biHeight;
   /* 
    printf("      fname: %s\n",fname);
    printf("         ID: %2d %2d\n",header.id1,header.id2);
    printf("   filesize: %d\n",header.bfSize);
    printf("data offset: %d\n",header.bfOffs);
    printf("             %d*%d*%d\n",header.biWidth,header.biHeight,header.biBitCnt);
    */
    if ((header.bfOffs-BITMAP_HEADER_SIZE>0) && (*bpp!=24)) {
      psize=(header.bfOffs-BITMAP_HEADER_SIZE)/4;
      if (*bpp==1)
	psize=2;
      palette=new RGB_QUAD [psize];
      if (*bpp!=1) {
	fseek(f,BITMAP_HEADER_SIZE,SEEK_SET);
	fread(palette,4,psize,f);
      } else {
	fread(palette,4,psize,f);
      }
      buffer=new byte [*w];
    }
    fseek(f,header.bfOffs,SEEK_SET);
    
    data=new RGBRec* [*h];
    for (i=*h-1;i>=0;i--) {
      data[i]=new RGBRec [*w];
      if (*bpp==1) {
	fread(buffer,1,*w/8,f);
	for (j=0;j<*w/8;j++) {
	  m=0x80;
	  for (k=0;k<8;k++) {
	    rgb=PixToRGB((buffer[j] & m)/m,palette);
	    data[i][j*8+k].Red  =rgb.Red;
	    data[i][j*8+k].Green=rgb.Green;
	    data[i][j*8+k].Blue =rgb.Blue;
	    m=m/2;
	  }
	}
      }
      if (*bpp==8) {
	fread(buffer,1,*w,f);
	for (j=0;j<*w;j++) {
	  rgb=PixToRGB(buffer[j],palette);
	  data[i][j].Red=rgb.Red;
	  data[i][j].Green=rgb.Green;
	  data[i][j].Blue=rgb.Blue;
	}
      }
      if (*bpp==24) {
	fread(data[i],3,*w,f);
      }
    }
    fclose(f);
    if (palette!=NULL) {
      delete palette;
      delete buffer;
    }
  } else
    data=NULL;

  return data;
}

byte *RGBRec2RGBField(byte **data,int w,int h) {
  byte *ret;
  int i;

  ret=new byte [3*w*h];
  for (i=0;i<h;i++)
    memcpy((ret+3*i*w),data[i],3*w);

  return ret;
}

void ExportToJpeg(int w,int h,int bpp,byte** data,char *name,int quality) {
  byte *image;
  unsigned int width,height;

  width=(unsigned int)w;
  height=(unsigned int)h;

  image=RGBRec2RGBField(data,w,h);
  JpegFile::BGRFromRGB(image,width,height);                                             
  JpegFile::RGBToJpegFile(name,image,width,height,1,quality);
  delete image;
}

void bmp2jpg(char *bmpname,char *jpgname,int quality) {
  UINT height,width;
  byte *image;
  int bpp;
  int qual;
  byte **bmpimage;

  bmpimage=(byte**)ReadBMP(bmpname,&width,&height,&bpp);
  image=RGBRec2RGBField(bmpimage,width,height);
  delete bmpimage;
  JpegFile::BGRFromRGB(image,width,height);                                             
  JpegFile::RGBToJpegFile(jpgname,image,width,height,1,quality);
  delete image;
}

